# -*- coding: utf-8 -*-
"""
Tools for interacting with Slack.

..codeauthor Grzegorz Pawełczuk <grzegorz.pawelczuk@nftlearning.com>
"""

from nftl_slack_tools.slack_api import SlackApi


class Chat(SlackApi):
    """
    Slack API wrapper that is handling chat category
    more info: https://api.slack.com/methods#chat
    """
    API_CATEGORY = "chat"

    def __init__(self, slack_token: str, slack_api_url: str = None) -> None:
        super().__init__(slack_token, slack_api_url)

    def delete(self, channel: str, ts: str, token: str = None) -> bool:
        """
        Deletes a message
        more info https://api.slack.com/methods/chat.delete

        Args:
            channel: Slack channel id like CXJSD234G
            ts: Message id
            token: optional auth token that will overwrite SlackApi token
        Returns:
            boolean operation status
        """

        params = {'channel': channel, 'ts': ts}
        call = super()._call(
            method='delete',
            data_key='channel',
            data=params,
            default=None,
            token=token,
            full_response=True
        )
        if call and len(call) > 0:
            return True
        return False



# -*- coding: utf-8 -*-
"""
Tools for interacting with Slack.

..codeauthor Grzegorz Pawełczuk <grzegorz.pawelczuk@nftlearning.com>
"""
import logging as log
from time import sleep
from typing import List, Optional, Dict, Any

from nftl_slack_tools.api.chat import Chat
from nftl_slack_tools.slack_api import SlackApi, STANDARD_DELAY, \
    THROTTLING_DELAY, SMALL_DELAY


class Channels(SlackApi):
    """
    Slack API wrapper that is handling users category
    more info https://api.slack.com/methods#channels
    """
    API_CATEGORY = "channels"

    def __init__(self, slack_token: str, slack_api_url: str = None) -> None:
        super().__init__(slack_token, slack_api_url)

    def list(self, limit: Optional[int] = 1000, cursor: Optional[str] = None,
             token: Optional[str] = None) -> Optional[Dict[str, Any]]:
        """
        Lists all channels in a Slack team.
        Channels depends on token so if you want to list users use user token
        instead of SlackApi admin token
        more info https://api.slack.com/methods/channels.list

        Args:
            limit: optional page length ( default: 1000 )
            cursor: optional message list cursor for pagination
            token: optional auth token that will overwrite SlackApi token
        Returns:
            None if data is invalid or channels :py:class:`typing.Dict`
            with two keys: channels (list of channels) and cursor for pagination
            when success
        """
        data: Dict[str, Any] = {'limit': limit}
        if cursor:
            data.update({'cursor': cursor})

        response = super()._call(
            method='list',
            data_key='',
            data=data,
            default={},
            token=token,
            http_get=True,
            full_response=True
        )

        if not response or len(response.get('channels', [])) < 1:
            return None

        return {
            'channels': response.get('channels', []),
            'cursor': response.get('response_metadata', {}).get('next_cursor')
        }

    def create(self, channel: str,
               token: Optional[str] = None) -> Optional[Dict[str, Any]]:
        """
        Create new channel with desired name
        more info https://api.slack.com/methods/channels.create

        Args:
            channel: channel name, can only contain lowercase letters, numbers,
                hyphens, and underscores, and must be 21 characters or less
            token: optional auth token that will overwrite SlackApi token
        Returns:
            None if data is invalid or channel info as :py:class:`typing.Dict`
            when success
        """
        params = {
            'name': channel,
            'validate': True
        }
        return super()._call(
            method='create',
            data_key='channel',
            data=params,
            default={},
            token=token
        )

    def onboard(self, users_ids: List, cursor: Optional[str] = None,
                invite_token: Optional[str] = None) -> bool:
        """
        Invite users to all the channels available in token scope except general

        Args:
            users_ids: :py:class:`typing.List` of slack user ids
            cursor: pagination cursor for retry etc.
            invite_token: optional auth token that will overwrite SlackApi token
                used for channel invitation
        Returns:
            boolean success status
        """
        channels = self.list(
            limit=10,
            cursor=cursor
        )
        if not channels:
            log.error('Channels are unavailable...')
            return False

        data = channels.get('channels', [])
        sleep(STANDARD_DELAY)

        for channel in data:
            channel_id = channel.get('id', None)
            is_archived = channel.get('is_archived', None)
            is_general = channel.get('is_general', None)

            if channel_id and not is_archived and not is_general:
                info = self.info(channel_id)
                sleep(STANDARD_DELAY)
                if not info:
                    log.warning("throttling break...")
                    sleep(THROTTLING_DELAY)
                    return self.onboard(users_ids, cursor, invite_token)

                members = info.get('members', [])
                for uid in users_ids:
                    if uid not in members:
                        log.debug(f'inviting "{uid}" into "{channel_id}"')
                        invite = self.invite(channel_id, uid, invite_token)
                        if not invite:
                            log.warning("throttling break...")
                            sleep(THROTTLING_DELAY)
                            return self.onboard(users_ids, cursor, invite_token)
                        sleep(STANDARD_DELAY)
                    else:
                        log.debug(f'invite aborted: '
                                  f'{uid}" already in "{channel_id}"')
            else:
                log.info(
                    f'channel skipped "{channel_id}": '
                    f'archived: {str(is_archived)}, general: {str(is_general)}'
                )

        if channels.get('cursor'):
            sleep(STANDARD_DELAY)
            return self.onboard(users_ids, channels.get('cursor'), invite_token)

        return True

    def invite(self, channel: str, user: str,
               token: Optional[str] = None) -> Optional[Dict[str, Any]]:
        """
        Invite user to the channel
        more info https://api.slack.com/methods/channels.invite

        Args:
            channel: Slack channel id like CXJSD234G
            user: Slack user id like UXS65F48
            token: optional auth token that will overwrite SlackApi token
        Returns:
            None if data is invalid or channel info as :py:class:`typing.Dict`
            when success
        """
        params = {'channel': channel, 'user': user}
        return super()._call(
            method='invite',
            data_key='channel',
            data=params,
            default={},
            token=token
        )

    def info(self, channel_id: str,
             token: Optional[str] = None) -> Optional[Dict[str, Any]]:
        """
        Get channel info by channel id
        more info https://api.slack.com/methods/channels.info

        Args:
            channel_id: Slack channel id
            token: optional auth token that will overwrite SlackApi token
        Returns:
            None if channel_id not found or channel info
            as :py:class:`typing.Dict` when success
        """
        params = {'channel': channel_id}
        return super()._call(
            method='info',
            data_key='channel',
            http_get=True,
            data=params,
            default={},
            token=token
        )

    def channel_info(self, channel_name: str,
                     token: Optional[str] = None) -> Optional[Dict[str, Any]]:
        """
        Find channel info by channel name
        There is no equivalent method in Slack API

        Args:
            channel_name: Slack channel name
            token: optional auth token that will overwrite SlackApi token
        Returns:
            None if channel_name is not found or channel info
            as :py:class:`typing.Dict` when success
        """
        channels = self.list(token=token, limit=1000)
        if not channels:
            log.error('Channels are unavailable')
            return None

        channel_id = ''
        channels_list = channels.get('channels', [])
        for channel in channels_list:
            name = channel.get('name')
            if name and name == channel_name and channel.get('id'):
                return channel
        else:
            log.warning(f'Channel "{channel_id}" not found')
            return None

    def history(self, channel: str, count: int = 100,
                latest: Optional[str] = None,
                token: Optional[str] = None) -> Optional[Dict[str, Any]]:
        """
        Fetches history of messages and events from a channel.
        more info https://api.slack.com/methods/channels.history

        Args:
            channel: Slack channel id like CXJSD234G
            count: number of messages per page
            latest: End of time range
            token: optional auth token that will overwrite SlackApi token
        Returns:
            None if error occurred or number of removed messages
        """
        params = {'channel': channel, 'count': f'{count}'}

        if latest:
            params['latest'] = latest

        return super()._call(
            method='history',
            data_key='',
            data=params,
            default=[],
            http_get=True,
            token=token,
            full_response=True
        )

    def clear_history(self, channel_name: str, chat: Chat) -> int:
        """
        Channel history cleaner
        Reads history and removes messages one by one

        Args:
            channel_name: Slack channel name
            chat: chat api handler
        Returns:
            Number of removed messages
        """
        deleted = 0
        channel = self.channel_info(channel_name)
        if channel:
            channel_id: str = channel.get('id', '')
            has_more = True
            msg_id = None
            log.info(f'channel id: "{channel_id}'"")

            while has_more:
                if msg_id:
                    history = self.history(channel_id, 10, msg_id)
                else:
                    history = self.history(channel_id, 10)
                sleep(STANDARD_DELAY)
                messages: list = []
                if history:
                    messages = history.get('messages', [])
                    has_more = history.get('has_more', False)

                for msg in messages:
                    msg_id = msg.get('ts')
                    log.info(f'deleting ts {msg_id}')
                    delete = chat.delete(channel_id, msg_id)
                    sleep(SMALL_DELAY)
                    if delete:
                        deleted += 1
                    else:
                        log.warning("throttling break...")
                        sleep(THROTTLING_DELAY)
                        return deleted + self.clear_history(channel_name, chat)

        return deleted



# -*- coding: utf-8 -*-
"""
Tools for interacting with Slack.

..codeauthor Grzegorz Pawełczuk <grzegorz.pawelczuk@nftlearning.com>
"""
import json
import requests
import logging as log

from abc import ABC
from typing import Any, Dict, Optional
from http.client import responses
from requests import Response, RequestException

STANDARD_DELAY = 1.1
SMALL_DELAY = 0.55
THROTTLING_DELAY = 5.05


class SlackApi(ABC):
    """
    Slack API abstract handler

    Args:
        slack_token: Slack admin auth token for API calls
        slack_api_url: Slack API base Url
    """

    API_CATEGORY = ""

    def __init__(self, slack_token: str, slack_api_url: Optional[str]) -> None:
        self.slack_api_url = slack_api_url
        if not slack_api_url:
            self.slack_api_url = 'https://slack.com/api/'
        self.slack_token = slack_token

    def _get_token(self, token: Optional[str]) -> str:
        """
        Validate token - use provided if available or provide admins one

        Args:
            token: optional users auth token
        Returns:
            Auth token that scan be used
        """
        return token if token else self.slack_token

    @staticmethod
    def _get_data(response) -> Optional[Dict[str, Any]]:
        """
        Universal Response validator

        Args:
            response - :py:class:`requests.Response` object to be validated
        Returns:
            None if data is invalid or Dict[str, Any] when success
        """
        if response.ok:
            data = response.json()
            if data.get('ok'):
                return response.json()
            error = data.get('error')
            if error:
                log.error(f'Slack API error {error} retrieving {response.url}')
        else:
            status = response.status_code
            name = responses[status]
            error = response.headers.get('x-amzn-errortype')
            msg = 'HTTP error (%s - %s, error: %s) retrieving: %s' % (
                status, name, error, response.url)
            log.error(msg)
        return None

    def _call(self, method: str, data_key: str, data: Dict[str, Any],
              default: Any, token: Optional[str] = None, http_get: bool = False,
              full_response: bool = False) -> Optional[Any]:
        """
        Wrapper for request builder

        Args:
            method: Slack API method name
            data_key: data key from response json that should be returned
            data: request parameters
            default: object that should be returned when data key
                will not be avilable
            token: optional users auth token
        Returns:
            data structure that is available in json or None
            when error occurred
        """
        req = self._call_slack(
            f'{self.API_CATEGORY}.%s' % method,
            data,
            token,
            http_get
        )

        if req is not None:
            resp_data = self._get_data(req)

            if full_response:
                return resp_data
            if resp_data:
                return resp_data.get(data_key, default)
        return None

    def _call_slack(self, method: str, data: Dict[str, Any],
                    token: Optional[str] = None,
                    http_get: bool = False) -> Optional[Response]:
        """
        Slack POST request builder

        Args:
            method: Slack API method name
            data: request parameters
            token: optional users auth token
        Returns:
            :py:class:`requests.Response` for provided request
        """
        url = f'{self.slack_api_url}{method}'
        try:
            if http_get:
                return requests.get(
                    url,
                    headers=self.get_headers(token),
                    params=data
                )
            return requests.post(
                url,
                headers=self.get_headers(token),
                data=bytes(json.dumps(data), 'utf-8')
            )
        except RequestException as e:
            log.error(f'Error handling request "{url}" with an exception: {e}')

        return None

    def get_headers(self, token: Optional[str] = None):
        return {
            'Authorization': 'Bearer %s' % self._get_token(token),
            'Content-type': 'application/json; charset=utf-8'
        }



# -*- coding: utf-8 -*-
"""
Tools for interacting with Slack.

..codeauthor Grzegorz Pawełczuk <grzegorz.pawelczuk@nftlearning.com>
"""
from typing import Optional, Dict, Any

from nftl_slack_tools.slack_api import SlackApi


class Users(SlackApi):
    """
    Slack API wrapper that is handling users category
    https://api.slack.com/methods#users
    """
    API_CATEGORY = "users"

    def __init__(self, slack_token: str, slack_api_url: str = None) -> None:
        super().__init__(slack_token, slack_api_url)

    def info(
            self,
            user: str,
            locale: bool = False,
            token: str = None) -> Optional[Dict[str, Any]]:
        """
        Gets information about the user
        more info https://api.slack.com/methods/users.info

        Args:
            user: Slack user id like UXS65F48
            locale: should response contain locale info
            token: optional auth token that will overwrite SlackApi token
        Returns:
            None if data is invalid or user info as :py:class:`typing.Dict`
            when success
        """
        include_locale = 0
        if locale:
            include_locale = 1
        params = {'user': user, 'include_locale': include_locale}
        return super()._call(
            method='info',
            data_key='user',
            data=params,
            default={},
            token=token,
            http_get=True
        )



# -*- coding: utf-8 -*-
"""
Tools for interacting with Slack.

..codeauthor Grzegorz Pawełczuk <grzegorz.pawelczuk@nftlearning.com>
"""
from typing import Optional

from nftl_slack_tools.api.channels import Channels
from nftl_slack_tools.api.chat import Chat
from nftl_slack_tools.api.users import Users


class SlackClient:
    """
    Slack API wrapper
    Partial implementation of methods that full documentation
    more info at  https://api.slack.com/methods
    """

    def __init__(self, slack_token: str, slack_api_url: str = None) -> None:
        self._token = slack_token
        self._slack_api_url = slack_api_url
        self._users_api: Optional[Users] = None
        self._channels_api: Optional[Channels] = None
        self._chat_api: Optional[Chat] = None

    def get_users_api(self) -> Users:
        """
        Provides Wrapper for methods available in Users category
        more info at https://api.slack.com/methods#users

        """
        if not self._users_api:
            self._users_api = Users(self._token, self._slack_api_url)
        return self._users_api

    def get_channels_api(self) -> Channels:
        """
        Provides Wrapper for methods available in Channels category
        more info at https://api.slack.com/methods#channels

        """
        if not self._channels_api:
            self._channels_api = Channels(self._token, self._slack_api_url)
        return self._channels_api

    def get_chat_api(self) -> Chat:
        """
        Provides Wrapper for methods available in Channels category
        more info at https://api.slack.com/methods#channels

        """
        if not self._chat_api:
            self._chat_api = Chat(self._token, self._slack_api_url)
        return self._chat_api



# -*- coding: utf-8 -*-
"""
Automated tests for Ed provisioner.

..codeauthor Grzegorz Pawełczuk <grzegorz.pawelczuk@nftlearning.com>
"""
import json

import pytest
import responses
from requests import HTTPError

from nftl_slack_tools.client import SlackClient


@pytest.fixture()
def delays(monkeypatch):
    monkeypatch.setattr("nftl_slack_tools.slack_api.STANDARD_DELAY", 0)
    monkeypatch.setattr("nftl_slack_tools.slack_api.SMALL_DELAY", 0)
    monkeypatch.setattr("nftl_slack_tools.slack_api.THROTTLING_DELAY", 0)


@pytest.fixture
def slack_client():
    return SlackClient('xoxp-xxxx')


@pytest.fixture
def broken_url_slack_client():
    return SlackClient('xoxp-xxxx', 'https://example.com/api')


def test_slack_users(broken_url_slack_client):
    # given:
    api1 = broken_url_slack_client.get_users_api()
    api2 = broken_url_slack_client.get_users_api()

    # then:
    assert api1 == api2


@responses.activate
def test_slack_users_info(slack_client):
    # given:
    responses.add(
        method=responses.GET,
        url='https://slack.com/api/users.info',
        body='{"ok":true,"user":{"test":"test"}}',
        status=200,
        content_type='application/json'
    )

    # then:
    assert slack_client.get_users_api().info('UXXX', locale=True)


@responses.activate
def test_slack_users_info_invalid(slack_client):
    # given:
    responses.add(
        method=responses.GET,
        url='https://slack.com/api/users.info',
        body='{"ok":false, "error":"error_key"}',
        status=200,
        content_type='application/json'
    )

    # then:
    assert not slack_client.get_users_api().info('UXXX', locale=True)


@responses.activate
def test_slack_users_info_invalid_network(slack_client):
    # given:
    responses.add(
        method=responses.GET,
        url='https://slack.com/api/users.info',
        body='',
        status=500,
        content_type='application/json'
    )

    # then:
    assert not slack_client.get_users_api().info('UXXX', locale=True)


def test_slack_channels(slack_client):
    # given:
    api1 = slack_client.get_channels_api()
    api2 = slack_client.get_channels_api()

    # then:
    assert api1 == api2


@responses.activate
def test_slack_channels_info(slack_client):
    # given:
    responses.add(
        method=responses.GET,
        url='https://slack.com/api/channels.list',
        body='{"ok":true,"channels":[{"some_channel_data":true}]}',
        status=200,
        content_type='application/json'
    )

    # then:
    data = slack_client.get_channels_api().list()
    assert isinstance(data, dict)
    assert isinstance(data.get('channels'), list)


@responses.activate
def test_slack_channels_create(slack_client):
    # given:
    responses.add(
        method=responses.POST,
        url='https://slack.com/api/channels.create',
        body='{"ok":true,"channel":{}}',
        status=200,
        content_type='application/json'
    )

    # then:
    assert isinstance(slack_client.get_channels_api().create('terefere'), dict)


@responses.activate
def test_slack_channels_invite(slack_client):
    # given:
    responses.add(
        method=responses.POST,
        url='https://slack.com/api/channels.invite',
        body='{"ok":true,"channel":{}}',
        status=200,
        content_type='application/json'
    )

    # then:
    assert isinstance(
        slack_client.get_channels_api().invite('terefere', 'UXXX'), dict)


@responses.activate
def test_slack_channels_channel_info_by_name_valid(slack_client):
    # given:
    _req_channel_info_by_name()

    # then:
    assert isinstance(
        slack_client.get_channels_api().channel_info('cname'), dict)


@responses.activate
def test_slack_channels_channel_info_exception(slack_client):
    # given:
    responses.add(responses.GET, 'https://slack.com/api/channels.list',
                  body=HTTPError('Error'))

    # then:
    assert not slack_client.get_channels_api().list(cursor='irrelevant')


@responses.activate
def test_slack_channels_channel_info_by_name_invalid_req(slack_client):
    # given:
    responses.add(
        method=responses.GET,
        url='https://slack.com/api/channels.list',
        body='',
        status=404,
        content_type='application/json'
    )

    # then:
    assert not slack_client.get_channels_api().list()


@responses.activate
def test_slack_channels_info_method(slack_client):
    # given:
    responses.add(
        method=responses.GET,
        url='https://slack.com/api/channels.info',
        body='',
        status=404,
        content_type='application/json'
    )

    # then:
    assert not slack_client.get_channels_api().info('irrelevant')


@responses.activate
def test_slack_channels_channel_info_by_name_invalid(slack_client):
    # given:
    responses.add(
        method=responses.GET,
        url='https://slack.com/api/channels.list',
        body='{"ok":false,"error":"boom!"}',
        status=200,
        content_type='application/json'
    )

    # then:
    assert not slack_client.get_channels_api().channel_info('cname')


@responses.activate
def test_slack_channels_channel_info_by_name_not_found(slack_client):
    # given:
    responses.add(
        method=responses.GET,
        url='https://slack.com/api/channels.list',
        body='{"ok":true,"channels":[{"id":"id","name":"boom"}]}',
        status=200,
        content_type='application/json'
    )

    # then:
    assert not slack_client.get_channels_api().channel_info('cname')


@responses.activate
def test_slack_channels_onboard_no_channels(slack_client, mocker, caplog):
    # given:
    mocker.patch(
        'nftl_slack_tools.api.channels.Channels.list').return_value = None

    # when:
    slack_client.get_channels_api().onboard(['irrelevant'])

    # then:
    assert 'Channels are unavailable' in caplog.text


@responses.activate
def test_slack_channels_onboard(slack_client, mocker, caplog, delays):
    # given:
    ret_present = mocker.patch('nftl_slack_tools.api.channels.Channels.list')
    channels_with_cursor = {'channels': [
        {'id': 'irrelevant', 'is_general': False, 'is_archived': False},
        {'id': 'irrelevant2', 'is_general': False, 'is_archived': False}],
        'cursor': 'next please'}
    channels_without_cursor = {'channels': [
        {'id': 'irrelevant', 'is_general': False, 'is_archived': False},
        {'id': 'irrelevant2', 'is_general': False, 'is_archived': True}]}

    ret_present.side_effect = [
        channels_with_cursor,
        channels_with_cursor,
        channels_with_cursor,
        channels_without_cursor
    ]
    ret_present = {
        'members': [
            'irrelevant'
        ]
    }
    ret_empty = {
        'members': []
    }
    info = mocker.patch('nftl_slack_tools.api.channels.Channels.info')
    info.side_effect = [None, ret_present, ret_empty, ret_present, ret_empty,
                        ret_present]

    invite = mocker.patch('nftl_slack_tools.api.channels.Channels.invite')
    invite.side_effect = [None, {'ok'}]

    # when:
    onboard = slack_client.get_channels_api().onboard(['irrelevant'])

    # then:
    assert 'throttling break' in caplog.text
    assert onboard


@responses.activate
def test_slack_channels_clear_history(slack_client, delays):
    def _response(has_more: bool = False):
        return {
            'ok': True,
            'has_more': has_more,
            'messages': [{
                'ts': '111',
                'name': channel_name
            }]
        }

    # given:
    channel_name = 'general'
    channel_id = 'xxx'
    api_ = 'https://slack.com/api'
    _req_channel_info_by_name(channel_name)

    responses.add(
        method=responses.GET,
        url='%s/channels.history?channel=%s&count=10'
            % (api_, channel_id),
        body=json.dumps(_response(True)),
        status=200,
        content_type='application/json'
    )

    responses.add(
        method=responses.GET,
        url='%s/channels.history?channel=%s&count=10&latest=%s'
            % (api_, channel_id, '111'),
        body=json.dumps(_response(False)),
        status=200,
        content_type='application/json'
    )

    _req_chat_delete(channel_id)
    _req_chat_delete(channel_id, 'false')
    _req_chat_delete(channel_id)

    # then:
    assert slack_client.get_channels_api().clear_history(
        channel_name,
        slack_client.get_chat_api()
    ) == 2


def _req_channel_info_by_name(name: str = 'cname'):
    responses.add(
        method=responses.GET,
        url='https://slack.com/api/channels.list',
        body='{"ok":true,"channels":[{"id":"xxx","name":"%s"}]}' % name,
        status=200,
        content_type='application/json'
    )


def _req_chat_delete(channel_id: str, ok: str = 'true',
                     base_url: str = 'https://slack.com/api'):
    responses.add(
        method=responses.POST,
        url=('%s/chat.delete' % base_url),
        body='{"ok":%s,"channel":"%s"}' % (ok, channel_id),
        status=200,
        content_type='application/json'
    )
